#include <stdlib.h>
#include <stdio.h>
#include <omp.h>

struct node {
   float val;
   struct node* l;
   struct node* r;
} nodeT;

#define MAXLEVEL 23

int lg2(int k) {
   int c = 0;
   while (k > 1) {
      k = k/2;
      c++;
   }
   return c;
}

struct node* build(int level) {

   if (level < MAXLEVEL) {
      struct node* p = (struct node*) malloc(sizeof(nodeT));
      p->val = (double)rand() / (double)((unsigned)RAND_MAX + 1);
      p->l = build(level+1);
      p->r = build(level+1);
      return p;
   } else {
      return NULL;
  }
}

// a sequential traversal of a subtree done by a task.
int traverse(struct node* p) {
   int c = 0;
   if (p == NULL) return;
   if (p->val < 0.5) c++;
   if (p->l == NULL) return 0;
   else {
      c+= traverse(p->l);
   }
   if (p->r == NULL) return 0;
   else {
      c+= traverse(p->r);
   }
   return c;
}

// a sequential traversal of the tree until lvl is 0
// capture the nodes to be traversed in parallel
int spawn(struct node* p, int lvl, int* nodeCount, struct node** nodes) {
   int c = 0;
   if (p == NULL) return 0;
   if (p->val < 0.5) c++;
   if (lvl == 0) {
      nodes[*nodeCount] = p->l;
      *nodeCount += 1;
      nodes[*nodeCount] = p->r;
      *nodeCount += 1;
      #pragma omp taskwait
   } else {
      c+= spawn(p->l, lvl-1, nodeCount, nodes);
      c+= spawn(p->r, lvl-1, nodeCount, nodes);
   }
   return c;
}

int main( ) {

   struct node* h = build(0);
   double t = -omp_get_wtime( );
   int SIZE = 16;
   int count[SIZE];
   struct node** nodes;
   int nodeCount;
   int i;

   omp_set_num_threads(SIZE);
   int l2 = lg2(SIZE); 
   nodes = (struct node**) malloc(SIZE*sizeof(struct node*));
   t = -omp_get_wtime( );
   count[0] = count[0] + spawn(h, l2-1, &nodeCount, nodes); 
   #pragma omp parallel for
   for (i = 0; i < SIZE; i++) {
      count[i] = traverse(nodes[i]);
   }
   for (i = 1; i < SIZE; i++) {
      count[0] += count[i];
   }
   

   t += omp_get_wtime( );
      
   printf("count: %d\n", count[0]);
   printf("time: %2f\n", t);
}
